<?php
/***************************************************** 
 *  Plugin name : responsive_iframe
 *  File name : index.php
 *  Summary : responsive iframe 
 *  Version   : 0.1
 *  Auther    : Karl Richard Lembach <http://cmsimple.lembach-kr.de>
 *  License : free
 *****************************************************
 * Useage : {{{responsive_iframe($string [,$max_width ,$min_width]); }}}
 * Example : {{{responsive_iframe('/ordner-im-root/datei.html',640,280);}}}
******************************************************/
/*
 * Prevent direct access.
 */
if (!defined('CMSIMPLE_XH_VERSION')) {
    header('HTTP/1.0 403 Forbidden');
    exit;
}
/*  Funktionsaufruf mit folgenden Parametern (Parameter 2 - 5 sind optional):
	1. Parameter: aufzurufende Datei inclusive dem Pfad. Wenn Verweis auf eigenen Server zeigt, den Pfad auf den Server-root beziehen.
	2. Parameter: Hindergrundfarbe des Frame, Wert als Hex-Zahl ohne #
	3. Parameter: Bordereigenschaften, String mit Angaben für Rahmendicke, Rahmenart und Rahmenfarbe -> '1px dotted #f00'
	4. Parameter: Maximale Breite, die das Frame einnehmen darf. Format: reine Zahl, Wert steht für px. Mindestwert = Mindestbreite.
	5. Parameter: Mindestbreite, die das Frame einnehmen soll. Format: reine Zahl, Wert steht für px. Mindestwert = 300.
*/

function responsive_iframe($string, $back_color = '#F8F8F8', $border_style = '1px solid #FFF', $max_width = 1920 ,$min_width = 300){

	$dec_back_color = hexdec($back_color);
	$back_color = dechex($dec_back_color);
	if (strlen($border_style) <= 2) {
		$border_style = '1px solid #FFF';
	}
	$max_width = intval($max_width);
	$min_width = intval($min_width);
	if ($min_width < 300) {
		$min_width = 300;
	}
	if ($max_width < $min_width) {
		$max_width = $min_width;
	}

	$map_wrapper =
		'max-width: ' . $max_width .'px; '.
		'min-width: ' . $min_width .'px; '.
		'max-height: 2048px; '.
		'margin: 0px auto; '.
		'padding: 0px; '.
		'border: ' . $border_style . '; '.
		'background-color: #' . $back_color . '; ';

    $framesrc =
		'src="'.
		$string.
		'"';

	$framestyle =
		'style ="'.
		'border: none; '.
		'width: 100%; '.
		'"';

	$framescript =
		'<script>
		function setIframeHeight() {
			var ifrm = document.getElementById("iframe");
			var doc = ifrm.contentDocument? ifrm.contentDocument: 
					ifrm.contentWindow.document;
			ifrm.style.visibility = "hidden";
			ifrm.style.height = "10px";
		
			doc = doc || document;
			var body = doc.body, html = doc.documentElement;
			var height = Math.max( body.scrollHeight, body.offsetHeight, 
					html.clientHeight, html.scrollHeight, html.offsetHeight );		
			ifrm.style.height = height + "px";
			ifrm.style.visibility = "visible";
			}'."\n".
		'</script>'."\n";

	$frame_string = 
		'<iframe id="iframe" title="'.
		$string.
		'" '.
		$framesrc.' '.
		$framestyle.' '.
		'allowfullscreen onload="setIframeHeight();">'."\n".
		'<p>Ihr Browser kann leider keine eingebetteten Frames anzeigen:<br>'."\n".
		'Sie können die eingebettete Seite über den folgenden Verweis aufrufen:<a href="'.
		$string.
		'">'.
		$string.
		'</a></p>'."\n".'</iframe>'."\n"; 

$output = 
	'<div style="'.
	$map_wrapper.
	'">'."\n".
	$framescript.
	$frame_string.
	'</div>';
return $output;

}
?>